#!/usr/bin/env python
# coding: utf-8

ABORTED = 'ABORTED'
COMMITTED = 'COMMITTED'
PURGED = 'PURGED'
PAUSED = 'PAUSED'
UNKNOWN = 'UNKNOWN'

STATUS = dict(ABORTED=ABORTED,
              COMMITTED=COMMITTED,
              PURGED=PURGED,
              )
